package cn.com.qmth.examcloud.core.oe.student.api.controller;

import cn.com.qmth.examcloud.api.commons.security.bean.User;
import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.core.oe.student.bean.*;
import cn.com.qmth.examcloud.core.oe.student.dao.FaceBiopsyItemRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.FaceBiopsyItemStepRepo;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.FaceBiopsyItemEntity;
import cn.com.qmth.examcloud.core.oe.student.dao.entity.FaceBiopsyItemStepEntity;
import cn.com.qmth.examcloud.core.oe.student.dao.enums.FaceBiopsyType;
import cn.com.qmth.examcloud.core.oe.student.service.*;
import cn.com.qmth.examcloud.support.Constants;
import cn.com.qmth.examcloud.support.enums.ExamRecordStatus;
import cn.com.qmth.examcloud.support.enums.FaceBiopsyScheme;
import cn.com.qmth.examcloud.support.enums.HandInExamType;
import cn.com.qmth.examcloud.support.examing.ExamRecordData;
import cn.com.qmth.examcloud.support.examing.ExamingHeartbeat;
import cn.com.qmth.examcloud.support.examing.ExamingSession;
import cn.com.qmth.examcloud.support.helper.FaceBiopsyHelper;
import cn.com.qmth.examcloud.support.redis.RedisKeyHelper;
import cn.com.qmth.examcloud.web.helpers.GlobalHelper;
import cn.com.qmth.examcloud.web.helpers.SequenceLockHelper;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import cn.com.qmth.examcloud.web.support.ControllerSupport;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.*;

import java.util.List;

/**
 * @Description 人脸活体检测接口
 * @Author lideyin
 * @Date 2019/10/14 10:59
 * @Version 1.0
 */
@Api(tags = "新人脸活体检测接口")
@RestController
@RequestMapping("${app.api.oe.student}/faceBiopsy")
public class FaceBiopsyController extends ControllerSupport {

    @Autowired
    private ExamingSessionService examingSessionService;

    @Autowired
    private FaceBiopsyService faceBiopsyService;

    @Autowired
    private FaceBiopsyItemRepo faceBiopsyItemRepo;

    @Autowired
    private FaceBiopsyItemStepRepo faceBiopsyItemStepRepo;

    @Autowired
    private ExamRecordDataService examRecordDataService;

    @Autowired
    private ExamFaceLivenessVerifyService examFaceLivenessVerifyService;

    @Autowired
    private ExamFaceLiveVerifyService examFaceLiveVerifyService;

    @Autowired
    private ExamControlService examControlService;

    @Autowired
    private RedisClient redisClient;

    @ApiOperation(value = "获取活体检测基本信息")
    @GetMapping("/getFaceBiopsyBaseInfo")
    public FaceBiopsyBaseInfo getFaceBiopsyBaseInfo(@RequestParam Long examRecordDataId) {
        User user = getAccessUser();
        Long studentId = user.getUserId();
        String sequenceLockKey = Constants.GET_FACE_BIOPSY_INFO_PREFIX + studentId;
        //系统在请求结束后会，自动释放锁，无需手动解锁
        SequenceLockHelper.getLock(sequenceLockKey);

        //判断考试记录id是否有效
        ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(examRecordDataId);
        if (examRecordData == null) {
            throw new StatusException("200101", "无效的考试记录");
        }

        if (ExamRecordStatus.EXAM_ING != examRecordData.getExamRecordStatus()) {
            throw new StatusException("200103", "考试已结束");
        }

        // 获取考试会话，判断考生是否已结束考试
        ExamingSession examSessionInfo = examingSessionService.getExamingSession(studentId);
        if (examSessionInfo == null) {
            throw new StatusException("200104", "考试会话已过期");
        }

        //考试未开启人脸活体检测，不允许获取活检信息
        Long rootOrgId = user.getRootOrgId();
        Long examId = examRecordData.getExamId();
        Long orgId = examRecordData.getOrgId();

        if (!FaceBiopsyHelper.isFaceVerify(rootOrgId, examId, studentId)) {
            throw new StatusException("200105", "本场考试未开启人脸活体检测");
        }

        FaceBiopsyScheme faceBiopsyScheme = FaceBiopsyHelper.getFaceBiopsyScheme(user.getRootOrgId());

        // 活检开始时间
        Integer faceVerifyMinute;

        if (FaceBiopsyScheme.FACE_CLIENT == faceBiopsyScheme) {
            // C端活体检测方案
            faceVerifyMinute = examFaceLiveVerifyService.calculateStartFaceVerifyMinute(examRecordDataId);
        } else if (FaceBiopsyScheme.FACE_MOTION == faceBiopsyScheme) {
            // Electron Client 自研活体检测方案
            faceVerifyMinute = faceBiopsyService.calculateFaceBiopsyStartMinute(examRecordDataId);
        } else {
            // FaceID活体检测方案
            String examingHeartbeatKey = RedisKeyHelper.getBuilder().examingHeartbeatKey(examSessionInfo.getExamRecordDataId());
            ExamingHeartbeat examingHeartbeat = redisClient.get(examingHeartbeatKey, ExamingHeartbeat.class);

            int usedMinute = null == examingHeartbeat ? 0 : examingHeartbeat.getCost().intValue() / 60;
            faceVerifyMinute = examFaceLivenessVerifyService.getFaceLivenessVerifyMinute(user.getRootOrgId(),
                    orgId, examId, studentId, examRecordData.getId(), usedMinute);
        }

        FaceBiopsyBaseInfo faceBiopsyBaseInfo = new FaceBiopsyBaseInfo();
        faceBiopsyBaseInfo.setIdentificationOfLivingBodyScheme(faceBiopsyScheme.getCode());
        faceBiopsyBaseInfo.setFaceVerifyMinute(faceVerifyMinute == null ? null : (faceVerifyMinute + 1));
        return faceBiopsyBaseInfo;
    }

    @ApiOperation(value = "获取人脸活体检测详细步骤")
    @GetMapping("/getFaceBiopsyInfo")
    public FaceBiopsyInfo getFaceBiopsyInfo(@RequestParam Long examRecordDataId) {
        User user = getAccessUser();
        Long studentId = user.getUserId();
        String sequenceLockKey = Constants.GET_FACE_BIOPSY_INFO_PREFIX + studentId;
        //系统在请求结束后会，自动释放锁，无需手动解锁
        SequenceLockHelper.getLock(sequenceLockKey);

        //判断考试记录id是否有效
        ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(examRecordDataId);
        if (examRecordData == null) {
            throw new StatusException("200101", "无效的考试记录");
        }

        if (ExamRecordStatus.EXAM_ING != examRecordData.getExamRecordStatus()) {
            throw new StatusException("200102", "考试已结束");
        }

        // 获取考试会话，判断考生是否已结束考试
        ExamingSession examSessionInfo = examingSessionService.getExamingSession(studentId);
        if (examSessionInfo == null) {
            throw new StatusException("200103", "考试会话已过期");
        }

        //考试未开启人脸活体检测，不允许获取活检信息
        Long rootOrgId = user.getRootOrgId();
        Long examId = examRecordData.getExamId();
        Long orgId = examRecordData.getOrgId();
        if (!FaceBiopsyHelper.isFaceVerify(rootOrgId, examId, studentId)) {
            throw new StatusException("200104", "本场考试未开启人脸活体检测");
        }

        return faceBiopsyService.getFaceBiopsyInfo(user.getRootOrgId(), examRecordDataId, FaceBiopsyType.FACE_MOTION);
    }


    @ApiOperation(value = "保存活体检测结果")
    @PostMapping("/saveFaceBiopsyResult")
    public SaveFaceBiopsyResultResp saveFaceBiopsyResult(@RequestBody SaveFaceBiopsyResultReq req) {
        User user = getAccessUser();
        Long studentId = user.getUserId();
        String sequenceLockKey = Constants.GET_FACE_BIOPSY_INFO_PREFIX + studentId;
        //系统在请求结束后会，自动释放锁，无需手动解锁
        SequenceLockHelper.getLock(sequenceLockKey);

        if (req.getExamRecordDataId() == null) {
            throw new StatusException("200104", "考试记录id不允许为空");
        }
        //判断考试记录id是否有效
        ExamRecordData examRecordData = examRecordDataService.getExamRecordDataCache(req.getExamRecordDataId());
        if (examRecordData == null) {
            throw new StatusException("200105", "无效的考试记录");
        }

        if (ExamRecordStatus.EXAM_ING != examRecordData.getExamRecordStatus()) {
            throw new StatusException("200105", "考试已结束");
        }

        if (req.getFaceBiopsyItemId() == null) {
            throw new StatusException("200106", "人脸活体检测明细id不允许为空");
        }
        FaceBiopsyItemEntity faceBiopsyItemEntity = GlobalHelper.getEntity(faceBiopsyItemRepo, req.getFaceBiopsyItemId(),
                FaceBiopsyItemEntity.class);
        if (faceBiopsyItemEntity == null) {
            throw new StatusException("200111", "人脸活体检测明细id不存在");
        }

        if (faceBiopsyItemEntity.getCompleted() == true) {
            throw new StatusException("200112", "不允许操作已结束的人脸活体检测数据");
        }

        if (req.getVerifySteps() == null || req.getVerifySteps().isEmpty()) {
            throw new StatusException("200107", "活体检测步骤不允许为空");
        }

        if (req.getVerifySteps().stream().anyMatch(p -> p.getStepId() == null)) {
            throw new StatusException("200108", "活体检测步骤id不允许为空");
        }

        if (req.getVerifySteps().stream().anyMatch(p -> p.getAction() == null)) {
            throw new StatusException("200109", "活体检测执行动作不允许为空");
        }

        if (!verifyStepsAllMatch(req.getFaceBiopsyItemId(), req.getExamRecordDataId(), req.getVerifySteps())) {
            throw new StatusException("200110", "活体检测步骤与原始定义不匹配");
        }
        SaveFaceBiopsyResultResp resp = faceBiopsyService.saveFaceBiopsyResult(req, studentId);

        //如果活检满足交卷条件，则系统自动交卷,自动交卷逻辑不应该影响活检保存结果，所以不能放一个事务中
        if (resp.getEndExam()) {
            examControlService.handInExam(req.getExamRecordDataId(), HandInExamType.AUTO, null, true);
        }
        return resp;
    }

    /**
     * 校验活检步骤和原始步骤是否匹配
     *
     * @param faceBiopsyItemId
     * @param verifySteps
     * @return
     */
    private boolean verifyStepsAllMatch(Long faceBiopsyItemId, Long examRecordDataId, List<FaceBiopsyStepInfo> verifySteps) {
        List<FaceBiopsyItemStepEntity> originalVerifySteps = faceBiopsyItemStepRepo.findByFaceBiopsyItemId(faceBiopsyItemId);

        if (originalVerifySteps == null || originalVerifySteps.isEmpty() ||
                originalVerifySteps.size() != verifySteps.size()) {
            return false;
        }

        for (int i = 0; i < originalVerifySteps.size(); i++) {
            FaceBiopsyItemStepEntity originalStep = originalVerifySteps.get(i);
            FaceBiopsyStepInfo newStep = verifySteps.get(i);
            //如果步骤id和动作以及考试记录id不同时匹配，则认为不匹配
            if (!(originalStep.getId().equals(newStep.getStepId()) &&
                    originalStep.getAction().equals(newStep.getAction()) &&
                    originalStep.getExamRecordDataId().equals(examRecordDataId))) {
                return false;
            }
        }
        return true;
    }

}
